package ira.export;

import ira.IRA;
import net.minecraft.client.Minecraft;
import net.minecraft.client.renderer.ItemModelMesher;
import net.minecraft.client.renderer.block.model.IBakedModel;
import net.minecraft.client.resources.*;
import net.minecraft.creativetab.CreativeTabs;
import net.minecraft.item.Item;
import net.minecraft.item.ItemStack;
import net.minecraft.util.NonNullList;
import net.minecraft.util.text.TextComponentString;
import net.minecraftforge.client.resource.SelectiveReloadStateHandler;
import net.minecraftforge.client.resource.VanillaResourceType;
import net.minecraftforge.fml.client.FMLClientHandler;
import net.minecraftforge.fml.common.registry.EntityEntry;
import net.minecraftforge.fml.common.registry.ForgeRegistries;
import roj.collect.FilterList;
import roj.collect.MyHashSet;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.lang.reflect.Field;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;
import java.util.function.Consumer;

import static ira.IRA.mc;

public class Exporter {
    private static final Set<ItemData> items = new MyHashSet<>(1000);
    private static final Set<EntityData> entities = new MyHashSet<>(500);

    private static final Object[] EMPTY = new Object[0];

    public static boolean allMode = false;

    public static String translate(String key) {
        return translate(key, EMPTY);
    }

    public static String translate(String key, Object... par) {
        return I18n.format(key, par);
    }

    private static void damageSearch(Item item, Consumer<ItemStack> consumer) {
        final ItemModelMesher mesher = IRA.itemRenderer.getItemModelMesher();

        MyHashSet<String> icons = new MyHashSet<>();

        ItemStack stack = new ItemStack(item, 1, 0);

        for (int i = 0; i < 16; i++) {
            try {
                stack.setItemDamage(i);
                IBakedModel model = mesher.getItemModel(stack);
                String name = stack.getDisplayName();

                if (icons.add(name + "@" + ((model == null) ? 0 : System.identityHashCode(model)))) {
                    consumer.accept(stack.copy());
                }
            } catch (Throwable ignored) {}
        }
    }

    @SuppressWarnings("unchecked")
    public static void reloadLang(Minecraft mc, Language language) {
        final LanguageManager lMan = mc.getLanguageManager();
        lMan.setCurrentLanguage(language);
        mc.gameSettings.language = language.getLanguageCode();

        try {
            List<IResourcePack> list = new ArrayList<>(10);

            Field field = FMLClientHandler.class.getDeclaredField("resourcePackList");
            field.setAccessible(true);

            List<IResourcePack> pack = (List<IResourcePack>) field.get(FMLClientHandler.instance());
            list.addAll(pack);

            for (ResourcePackRepository.Entry entry : mc.getResourcePackRepository().getRepositoryEntries()) {
                list.add(entry.getResourcePack());
            }
            lMan.parseLanguageMetadata(list);

            SelectiveReloadStateHandler.INSTANCE.beginReload((t) -> t == VanillaResourceType.LANGUAGES);
            try {
                ((SimpleReloadableResourceManager) mc.getResourceManager()).notifyReloadListeners();
            } finally {
                SelectiveReloadStateHandler.INSTANCE.endReload();
            }

            sendMsg("使用快速方法完成更换语言!");
        } catch (Throwable e) {
            FMLClientHandler.instance().refreshResources(VanillaResourceType.LANGUAGES);
            e.printStackTrace();
        }

        mc.fontRenderer.setUnicodeFlag(lMan.isCurrentLocaleUnicode() || mc.gameSettings.forceUnicodeFont);
        mc.fontRenderer.setBidiFlag(lMan.isCurrentLanguageBidirectional());

        //mc.gameSettings.sendSettingsToServer();
    }

    private static void sendMsg(String s) {
        mc.player.sendMessage(new TextComponentString(s));
    }

    public static void exportMods(String modId) throws IOException {
        long start = System.currentTimeMillis();
        long tmp = start, tmp2;

        final Set<ItemData> items = Exporter.items;
        final Set<EntityData> entities = Exporter.entities;

        Language lang = mc.getLanguageManager().getCurrentLanguage();

        if(!lang.getLanguageCode().equalsIgnoreCase("zh_cn"))
            reloadLang(mc, mc.getLanguageManager().getLanguage("zh_cn"));

        sendMsg("更换中文语言: " + ((tmp2 = System.currentTimeMillis()) - tmp) + "ms");

        NonNullList<ItemStack> itemProcessor = new NonNullList<>(new FilterList<>((old, stack) -> {
            if(!IRA.blacklist.isEmpty()) {
                if (IRA.blacklist.contains(stack.getItem().getRegistryName().toString() + "@" + stack.getMetadata()))
                    return false;
            }

            final ItemData data = new ItemData(stack);

            data.setName(stack.getDisplayName());
            data.setCreativeName(getCreativeTabName(data));

            items.add(data);

            return false;
        }), ItemStack.EMPTY);

        for (Item item : Item.REGISTRY) {
            if (item.getRegistryName().getNamespace().equals(modId)) {
                final CreativeTabs tab = item.getCreativeTab();
                item.getSubItems(tab == null ? CreativeTabs.SEARCH : tab, itemProcessor);
                if(allMode) {
                    damageSearch(item, itemProcessor::add);
                }
            }
        }

        for (EntityEntry entity : ForgeRegistries.ENTITIES) {
            if (entity.getRegistryName().getNamespace().equals(modId)) {
                final EntityData data = new EntityData(entity);
                data.setName(translate("entity." + data.getEntity().getName() + ".name"));
                entities.add(data);
            }
        }

        sendMsg("中文 + 渲染: " + ((tmp = System.currentTimeMillis()) - tmp2) + "ms");

        reloadLang(mc, mc.getLanguageManager().getLanguage("en_us"));

        for (ItemData data : items) {
            data.setEnglishName(data.getItemStack().getDisplayName());
        }

        for (EntityData data : entities) {
            data.setEnglishName(translate("entity." + data.getEntity().getName() + ".name"));
        }

        sendMsg("添加英文名: " + ((tmp2 = System.currentTimeMillis()) - tmp) + "ms");

        File export = new File(mc.gameDir, "export/");
        if (!export.isDirectory() && !export.mkdirs()) {
            throw new IOException("无法创建导出目录");
        }

        modId = IRA.FILE_NAME_MATCHER.reset(modId).replaceAll("");

        File itemJson = new File(mc.gameDir, "export/" + modId + "_item.json");

        StringBuilder sb = new StringBuilder(1000000).append("{\"//\": \"Powered by IRAR(Iten Render Async Reloaded)\"}\n");
        for (ItemData data : items) {
            data.toJSON(sb).append('\n');
        }
        try (FileOutputStream fos = new FileOutputStream(itemJson)) {
            fos.write(sb.toString().getBytes(StandardCharsets.UTF_8));
        }

        File entityJson = new File(mc.gameDir, "export/" + modId + "_entity.json");

        deleteAll(sb);

        sb.append("{\"//\": \"Powered by IRAR(Iten Render Async Reloaded)\"}\n");
        for (EntityData data : entities) {
            data.toJSON(sb).append('\n');
        }
        try (FileOutputStream fos = new FileOutputStream(entityJson)) {
            fos.write(sb.toString().getBytes(StandardCharsets.UTF_8));
        }

        sendMsg("导出文件: " + (System.currentTimeMillis() - tmp2) + "ms");

        if(!lang.getLanguageCode().equalsIgnoreCase("en_us"))
            reloadLang(mc, lang);

        items.clear();
        entities.clear();

        sendMsg("Total: " + (System.currentTimeMillis() - start) + "ms");
    }

    static Field abstFb_count;
    static boolean isOracleVM;
    private static void deleteAll(StringBuilder sb) {
        if(isOracleVM) {
            if (abstFb_count == null) {
                try {
                    abstFb_count = StringBuilder.class.getSuperclass().getDeclaredField("count");
                    abstFb_count.setAccessible(true);
                } catch (NoSuchFieldException e) {
                    isOracleVM = false;
                    e.printStackTrace();
                }
            }

            try {
                abstFb_count.setInt(sb, 0);
            } catch (IllegalAccessException e) {
                isOracleVM = false;
                e.printStackTrace();
            }
        } else {
            sb.delete(0, sb.length());
        }
    }

    public static String getCreativeTabName(ItemData data) {
        final CreativeTabs tab = data.getItemStack().getItem().getCreativeTab();
        if (tab != null)
            return translate(tab.getTranslationKey());
        return "";
    }
}
