package ira.export;

import ira.IRA;
import ira.rendering.FrameBufWrapper;
import ira.rendering.Renderer;
import net.minecraft.item.ItemStack;
import net.minecraftforge.oredict.OreDictionary;
import roj.collect.MyHashMap;
import roj.config.data.CMapping;

import java.util.Arrays;
import java.util.List;
import java.util.Objects;

public class ItemData {
    private static final FrameBufWrapper fboLarge = new FrameBufWrapper(128),
            fboSmall = new FrameBufWrapper(32);

    private String name, engName, oredicts, tabName;

    private final String regName, type, smallIcon, largeIcon;

    private final int meta, maxStack, maxDurability;

    private final ItemStack stack;

    public ItemData(ItemStack stack) {
        this.regName = stack.getItem().getRegistryName().toString();
        this.meta = stack.getMetadata();

        if (!stack.isEmpty()) {
            final int[] ids = OreDictionary.getOreIDs(stack);
            if(ids.length > 0) {
                List<String> list = Arrays.asList(new String[ids.length]);
                for (int j = 0; j < ids.length; j++) {
                    list.set(j, OreDictionary.getOreName(ids[j]));
                }
                this.oredicts = list.toString();
            } else {
                this.oredicts = "[]";
            }
        }

        this.type = IRA.B2I.containsValue(stack.getItem()) ? "Block" : "Item";

        this.maxStack = stack.getMaxStackSize();
        this.maxDurability = stack.getMaxDamage() + 1;

        this.smallIcon = Renderer.getItemBase64(stack, fboSmall, IRA.itemRenderer);
        this.largeIcon = Renderer.getItemBase64(stack, fboLarge, IRA.itemRenderer);
        this.stack = stack;
    }

    public ItemStack getItemStack() {
        return this.stack;
    }

    public void setCreativeName(String name) {
        this.tabName = name;
    }

    public void setName(String name) {
        this.name = name;
    }

    public void setEnglishName(String englishName) {
        this.engName = englishName;
    }

    static CMapping map = new CMapping(new MyHashMap<>());

    public StringBuilder toJSON(StringBuilder sb) {
        final CMapping map = ItemData.map;

        map.clear();

        map.put("name", name);
        map.put("englishName", engName);
        map.put("registerName", regName);
        map.put("metadata", meta);
        map.put("OredictList", oredicts);
        map.put("CreativeTabName", tabName);
        map.put("type", type);
        map.put("smallIcon", smallIcon);
        map.put("largeIcon", largeIcon);
        map.put("maxStackSize", maxStack);
        map.put("maxDurability", maxDurability);

        return map.toShortJSON(sb);
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        ItemData itemData = (ItemData) o;

        return stack.getItem() == itemData.stack.getItem() && stack.getItemDamage() == itemData.stack.getItemDamage() && Objects.equals(itemData.stack.getTagCompound(), stack.getTagCompound());
    }

    @Override
    public int hashCode() {
        return stack.hashCode();
    }
}
