package ira.rendering;

import ira.util.PNGUtil;
import net.minecraft.client.Minecraft;
import net.minecraft.client.entity.EntityPlayerSP;
import net.minecraft.client.renderer.GLAllocation;
import net.minecraft.client.renderer.GlStateManager;
import net.minecraft.client.renderer.OpenGlHelper;
import net.minecraft.util.text.TextComponentString;
import org.apache.commons.codec.binary.Base64;
import org.lwjgl.BufferUtils;
import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL12;
import org.lwjgl.opengl.GL30;
import org.lwjgl.util.glu.GLU;
import roj.util.ByteList;

import javax.annotation.Nonnull;
import javax.imageio.ImageIO;
import java.awt.image.BufferedImage;
import java.awt.image.DataBufferInt;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.nio.ByteBuffer;
import java.nio.IntBuffer;

import static org.lwjgl.opengl.GL11.*;

public final class FrameBufWrapper {
    private int texSize;
    private int idFrameBuf = -1,
            idDepthBuf = -1,
            idTexture = -1;

    private final IntBuffer lastViewport = GLAllocation.createDirectIntBuffer(16);

    private int idTextureLast, idFrameBufLast;

    public FrameBufWrapper(int size) {
        this.texSize = size;
        createFBO();
    }

    public static void checkGlErrors(String message) {
        int error = GL11.glGetError();
        if (error != 0) {
            String name = GLU.gluErrorString(error);

            EntityPlayerSP player = Minecraft.getMinecraft().player;

            player.sendMessage(new TextComponentString("\u00a7e########## GL ERROR ##########"));
            player.sendMessage(new TextComponentString("\u00a7e@ " + message));
            player.sendMessage(new TextComponentString("\u00a7e" + error + ": " + name));
        }
    }

    public void resize(int newSize) {
        deleteFBO();
        this.texSize = newSize;
        createFBO();
    }

    public void begin() {
        checkGlErrors("FBO Begin Init");

        this.idFrameBufLast = GL11.glGetInteger(GL30.GL_FRAMEBUFFER_BINDING);

        OpenGlHelper.glBindFramebuffer(GL30.GL_FRAMEBUFFER, this.idFrameBuf);

        this.lastViewport.clear();
        GL11.glGetInteger(GL_VIEWPORT, this.lastViewport);
        GL11.glViewport(0, 0, this.texSize, this.texSize);

        GlStateManager.matrixMode(GL_MODELVIEW);

        GlStateManager.pushMatrix();
        GlStateManager.loadIdentity();
        this.idTextureLast = GL11.glGetInteger(GL_TEXTURE_BINDING_2D);

        GlStateManager.clearColor(0, 0, 0, 0);
        GlStateManager.clear(16640);

        GlStateManager.cullFace(GlStateManager.CullFace.FRONT);

        GlStateManager.enableDepth();
        GlStateManager.enableLighting();
        GlStateManager.enableRescaleNormal();

        checkGlErrors("FBO Begin Final");
    }

    public void end() {
        checkGlErrors("FBO End Init");

        GlStateManager.cullFace(GlStateManager.CullFace.BACK);
        GlStateManager.disableDepth();
        GlStateManager.disableRescaleNormal();
        GlStateManager.disableLighting();
        GlStateManager.matrixMode(GL_MODELVIEW);
        GlStateManager.popMatrix();
        GL11.glViewport(this.lastViewport.get(0), this.lastViewport.get(1), this.lastViewport.get(2), this.lastViewport.get(3));

        OpenGlHelper.glBindFramebuffer(GL30.GL_FRAMEBUFFER, this.idFrameBufLast);
        GlStateManager.bindTexture(this.idTextureLast);

        checkGlErrors("FBO End Final");
    }

    public void bind() {
        GlStateManager.bindTexture(this.idTexture);
    }

    void restoreTexture() {
        GlStateManager.bindTexture(this.idTextureLast);
    }

    public void saveToFile(File file) {
        BufferedImage image = getImage();

        File father = file.getParentFile();
        if(!father.isDirectory() && !father.mkdirs())
            throw new IllegalStateException("Cant create dir");

        try {
            PNGUtil.writeImage(image, new FileOutputStream(file));
        } catch (Exception e) {
            throw new IllegalStateException("Cant save png", e);
        }
    }

    public String getBase64() {
        BufferedImage image = getImage();

        OutputStream out = byteList.asOutputStream();

        try {
            PNGUtil.writeImage(image, out);
            String result = base64.encodeToString(byteList.toByteArray());
            byteList.clear();
            return result;
        } catch (IOException e) {
            throw new IllegalStateException("Cant save png", e);
        }
    }

    @Nonnull
    private BufferedImage getImage() {
        GlStateManager.bindTexture(this.idTexture);

        GL11.glPixelStorei(GL_PACK_ALIGNMENT, 1);
        GL11.glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

        int w = GL11.glGetTexLevelParameteri(GL_TEXTURE_2D, 0, GL_TEXTURE_WIDTH),
                h = GL11.glGetTexLevelParameteri(GL_TEXTURE_2D, 0, GL_TEXTURE_HEIGHT);

        IntBuffer texture = createIntBuffer(w * h);

        GL11.glGetTexImage(GL_TEXTURE_2D, 0, /*GL_RGBA*/GL12.GL_BGRA, GL12.GL_UNSIGNED_INT_8_8_8_8_REV, texture);

        BufferedImage image = new BufferedImage(this.texSize, this.texSize, BufferedImage.TYPE_INT_ARGB);

        int[] buffer = ((DataBufferInt)image.getRaster().getDataBuffer()).getData();

        texture.get(buffer);
        return image;
    }

    static Base64 base64 = new Base64(0, new byte[] {13, 10}, false);

    static ByteList byteList = new ByteList(8192);

    static IntBuffer currentBuffer;
    private static IntBuffer createIntBuffer(int size) {
        if(currentBuffer == null || currentBuffer.capacity() < size) {
            currentBuffer = null; // Help GC
            return currentBuffer = BufferUtils.createIntBuffer(size);
        } else {
            currentBuffer.clear();
            return currentBuffer;
        }
    }

    private void createFBO() {
        if(this.idFrameBuf != -1) {
            deleteFBO();
        }

        final int currFBO = GL11.glGetInteger(GL30.GL_FRAMEBUFFER_BINDING),
                currTex = GL11.glGetInteger(GL_TEXTURE_BINDING_2D);

        OpenGlHelper.glBindFramebuffer(GL30.GL_FRAMEBUFFER, this.idFrameBuf = OpenGlHelper.glGenFramebuffers());

        GL11.glBindTexture(GL_TEXTURE_2D, this.idTexture = GL11.glGenTextures());
        GL11.glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        GL11.glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        GL11.glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL12.GL_CLAMP_TO_EDGE);
        GL11.glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL12.GL_CLAMP_TO_EDGE);
        GL11.glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, this.texSize, this.texSize, 0, GL12.GL_BGRA, GL_UNSIGNED_BYTE, (ByteBuffer) null);

        GL11.glBindTexture(GL_TEXTURE_2D, currTex);

        OpenGlHelper.glBindRenderbuffer(GL30.GL_RENDERBUFFER, this.idDepthBuf = OpenGlHelper.glGenRenderbuffers());
        OpenGlHelper.glRenderbufferStorage(GL30.GL_RENDERBUFFER, GL_DEPTH_COMPONENT, this.texSize, this.texSize);
        OpenGlHelper.glFramebufferRenderbuffer(GL30.GL_FRAMEBUFFER, GL30.GL_DEPTH_ATTACHMENT, GL30.GL_RENDERBUFFER, this.idDepthBuf);
        OpenGlHelper.glFramebufferTexture2D(GL30.GL_FRAMEBUFFER, GL30.GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, this.idTexture, 0);

        OpenGlHelper.glBindFramebuffer(GL30.GL_FRAMEBUFFER, currFBO);
    }

    @Override
    protected void finalize() {
        deleteFBO();
    }

    private void deleteFBO() {
        if(this.idFrameBuf != -1) {
            OpenGlHelper.glDeleteFramebuffers(this.idFrameBuf);
            GL11.glDeleteTextures(this.idTexture);
            OpenGlHelper.glDeleteRenderbuffers(this.idDepthBuf);

            this.idDepthBuf = -1;
            this.idFrameBuf = -1;
            this.idTexture = -1;
        }
    }
}
