package ira.rendering;

import ira.IRA;
import net.minecraft.client.renderer.GlStateManager;
import net.minecraft.client.renderer.OpenGlHelper;
import net.minecraft.client.renderer.RenderHelper;
import net.minecraft.client.renderer.RenderItem;
import net.minecraft.client.renderer.entity.RenderManager;
import net.minecraft.entity.Entity;
import net.minecraft.entity.EntityList;
import net.minecraft.entity.EntityLivingBase;
import net.minecraft.item.ItemStack;
import net.minecraft.util.math.AxisAlignedBB;
import net.minecraftforge.fml.common.registry.EntityEntry;
import org.lwjgl.opengl.GL11;

import java.io.File;

import static ira.IRA.mc;
import static org.lwjgl.opengl.GL11.GL_MODELVIEW;
import static org.lwjgl.opengl.GL11.GL_PROJECTION;

public class Renderer {
    public static final float DEGREE_30 = (float) Math.toDegrees(Math.asin(Math.tan(Math.toRadians(30))));
    public static final float TAN0_25 = (float) Math.atan(0.025) * 20;

    public static void renderEntity(EntityLivingBase entity, FrameBufWrapper fbo, String suffix, boolean renderPlayer) {
        renderEntity(fbo, renderPlayer, entity);

        String name = EntityList.getEntityString(entity);
        if(name == null)
            name = entity.getName();

        fbo.saveToFile(new File(mc.gameDir, renderPlayer ? "rendered/player.png" : "rendered/entity_" + IRA.FILE_NAME_MATCHER.reset(name).replaceAll("") + suffix + ".png"));
        fbo.restoreTexture();
    }

    public static void renderItem(ItemStack stack, FrameBufWrapper fbo, String suffix, RenderItem ir) {
        renderItem(stack, fbo, ir);
        fbo.saveToFile(new File(mc.gameDir, "rendered/item_" + IRA.FILE_NAME_MATCHER.reset(stack.getItem().getTranslationKey()).replaceAll("") + stack.getItemDamage() + suffix + ".png"));
        fbo.restoreTexture();
    }

    public static String getItemBase64(ItemStack stack, FrameBufWrapper fbo, RenderItem ir) {
        renderItem(stack, fbo, ir);
        String base64 = fbo.getBase64();
        fbo.restoreTexture();
        return base64;
    }

    public static String getEntityBase64(EntityEntry entry, FrameBufWrapper fbo) {
        Entity entity = entry.newInstance(mc.world);
        if(!(entity instanceof EntityLivingBase))
            return "";
        renderEntity(fbo, false, (EntityLivingBase) entity);
        String base64 = fbo.getBase64();
        fbo.restoreTexture();
        return base64;
    }

    private static void renderItem(ItemStack stack, FrameBufWrapper fbo, RenderItem itemRenderer) {
        fbo.begin();
        GlStateManager.matrixMode(GL_PROJECTION);
        GlStateManager.pushMatrix();
        GlStateManager.loadIdentity();
        GlStateManager.ortho(0, 16, 0, 16, -150, 150);
        GlStateManager.matrixMode(GL_MODELVIEW);
        RenderHelper.enableGUIStandardItemLighting();
        GlStateManager.enableRescaleNormal();
        GlStateManager.enableColorMaterial();
        GlStateManager.enableLighting();
        float scale = IRA.renderScale;
        GlStateManager.translate(8 * (1 - scale), 8 * (1 - scale), 0);
        GlStateManager.scale(scale, scale, scale);
        itemRenderer.renderItemIntoGUI(stack, 0, 0);
        GlStateManager.disableLighting();
        RenderHelper.disableStandardItemLighting();
        GlStateManager.matrixMode(GL_PROJECTION);
        GL11.glPopMatrix();
        fbo.end();
    }

    private static void renderEntity(FrameBufWrapper fbo, boolean renderPlayer, EntityLivingBase entity) {
        fbo.begin();

        AxisAlignedBB aabb = entity.getEntityBoundingBox();
        final double boundLimit = Math.max(
                        Math.abs(Math.min(
                                aabb.minX - entity.posX, Math.min(
                                aabb.minY - entity.posY,
                                aabb.minZ - entity.posZ))),
                        Math.abs(Math.max(
                                aabb.maxX - entity.posX, Math.max(
                                aabb.maxY - entity.posY,
                                aabb.maxZ - entity.posZ))));

        GlStateManager.matrixMode(GL_PROJECTION);
        GlStateManager.pushMatrix();

        GlStateManager.loadIdentity();
        GlStateManager.ortho(-boundLimit * 0.75D, boundLimit * 0.75D, -boundLimit * 1.25D, boundLimit * 0.25D, -100, 100);

        GlStateManager.matrixMode(GL_MODELVIEW);
        GlStateManager.enableColorMaterial();
        GlStateManager.pushMatrix();
        GlStateManager.translate(0, 0, 50);

        if (renderPlayer) {
            GlStateManager.scale(-1, 1, 1);
        } else {
            float scale = IRA.renderScale;
            GlStateManager.scale(-scale, scale, scale);
        }

        GlStateManager.rotate(180, 0, 0, 1);

        float bak1 = entity.renderYawOffset,
                bak2 = entity.rotationYaw,
                bak3 = entity.rotationPitch,
                bak4 = entity.prevRotationYawHead,
                bak5 = entity.rotationYawHead;

        GlStateManager.rotate(135, 0, 1, 0);
        RenderHelper.enableStandardItemLighting();
        GlStateManager.rotate(-135, 0, 1, 0);

        GlStateManager.rotate(DEGREE_30, 1, 0, 0);
        GlStateManager.rotate(-45, 0, 1, 0);

        entity.renderYawOffset = TAN0_25;
        entity.rotationYaw = TAN0_25 * 2;
        entity.rotationPitch = -TAN0_25;

        entity.rotationYawHead = entity.rotationYaw;
        entity.prevRotationYawHead = entity.rotationYaw;

        GlStateManager.translate(0, 0, 0);
        RenderManager man = mc.getRenderManager();
        man.setPlayerViewY(180);
        man.setRenderShadow(false);
        man.renderEntity(entity, 0, 0, 0, 0, 1, true);
        man.setRenderShadow(true);

        entity.renderYawOffset = bak1;
        entity.rotationYaw = bak2;
        entity.rotationPitch = bak3;
        entity.prevRotationYawHead = bak4;
        entity.rotationYawHead = bak5;

        GlStateManager.popMatrix();

        RenderHelper.disableStandardItemLighting();
        GlStateManager.disableRescaleNormal();
        GlStateManager.setActiveTexture(OpenGlHelper.lightmapTexUnit);
        GlStateManager.disableTexture2D();
        GlStateManager.setActiveTexture(OpenGlHelper.defaultTexUnit);

        GlStateManager.matrixMode(GL_PROJECTION);
        GlStateManager.popMatrix();

        fbo.end();
    }
}
